/* Copyright (C) 2014-2017 Andrej Y. Rzhavskov aka rjaan

   single_llist.h  holds C-prototypes of single linked list

   This program is free software; the Free Software Foundation
   gives unlimited permission to copy, distribute and modify it.  
*/
#ifndef __SINGLE_LLIST_H__
#define __SINGLE_LLIST_H__	  1

#ifdef __cplusplus 
extern "C" {
#endif /*__cplusplus*/ 

#include <stdlib.h> 
#include <stdio.h>

struct __single_llist 
  { 
     struct __single_llist *next; /* a pointer on next element or NULL is end of list */
     void*       data;            /* some data */
     size_t      size;            /* the data size but it will equal 0 */
                                  /* when data contain a pointer address */   
  };    

typedef struct __single_llist  single_ll_t;

#define SINGLE_LLIST_LENGTH     sizeof(single_ll_t)
#define SINGLE_LLIST_ITEM_NEW(buf,size) ({                              	\
   single_ll_t  *__x = (single_ll_t  *)calloc(1,SINGLE_LLIST_LENGTH);    	\
   size_t (__size) = (size); char* __src = (char *)(buf);               	\
   do                								\
    {                                                                   	\
       if( (__x) ) {                                                      \
        if( (__size) > 0 && ((__x)->data=calloc( 1, ((__x)->size=(__size)))) )  \
         {                                                                      \
           char* __dst = (char*)((__x)->data);                                  \
           while ( (__src) && ((__src) - (char *)(buf)) < (__x)->size )         \
               *(__dst)++ = *(__src)++;                                         \
         }								        \
        else{                                                                   \
             (__x)->data=(void *)buf;                                           \
        }                                                               	\
      }                                                                  	\
   }										\
  while(0); __x; })

 static __inline__ single_ll_t* single_llist_item_new ( void *buf_p, size_t size )
 {
   return SINGLE_LLIST_ITEM_NEW(buf_p,size); 
 } 

 #define SINGLE_LLIST_ITEM_FREE(x) ({                                    \
    single_ll_t  *__x = (x);                                             \
    if ( (__x) ) {                                                       \
        if ( (__x)->size > 0 && (__x)->data ) { free((__x)->data); (__x)->data = NULL; }    \
           free((__x)); (__x) = NULL;                                        \
    }                                                                    \
     __x;})                      

 static __inline__ void single_llist_item_free ( single_ll_t **x_pp )
 {  
   if ( x_pp )
      *x_pp = SINGLE_LLIST_ITEM_FREE(*x_pp);
 }

 #define SINGLE_LLIST_ITEM_APPEND(L,x) ({                                \
    single_ll_t  *__head = (L), *__x = (x), *__prev=NULL;                \
    if ( x )                                                             \
      do{                                                                \
        for ( ; (__head) ; (__head)= (__prev)->next )                    \
                 (__prev) = (__head);                                    \
        if ( (__x) ) {                                                   \
           (__head) = (__x);                                             \
           if ( (__prev) ) (__prev)->next = (__head);                    \
         }                                                                \
        }while(0); __head; })                                                

 static __inline__ single_ll_t  *single_llist_item_append ( single_ll_t *list_p, single_ll_t *x_p )      
 {
    single_ll_t  *head_p = SINGLE_LLIST_ITEM_APPEND(list_p,x_p);
    return ( !list_p ? head_p : list_p );   
 }

 static __inline__ single_ll_t  *single_llist_item_pop(single_ll_t ** list_pp) {  
     single_ll_t  *item_p = ( list_pp ? *list_pp : NULL );
     if( item_p ) {
       *list_pp = item_p->next;         
     }  
     return item_p;
 }  

 #define SINGLE_LLIST_ITEM_PUSH(L,x) ({                                  \
    single_ll_t  *__head = (L), *__x = (x);                             \
     if ( (__x) ) {                                                      \
        (__x)->next = (__head);                                          \
        (__head) = (__x);                                                \
     }                                                                   \
     __head; })

 #define SINGLE_LLIST_DEPTH(L) ({                                        \
     single_ll_t  *__head = (L); size_t __n = 0;                         \
     do{                                                                   \
       while ( (__head) != NULL ) { (__head) = (__head)->next; (__n)++; }  \
     }while(0); __n; })

 #define SINGLE_LLIST_ITEM_GET(L,idx) ({                                 \
   single_ll_t  *__head = (L);                                          \
   unsigned long __idx = (idx), __n = 0;                                \
   do{                                                                  \
       for ( ; (__head) ; (__head)=(__head)->next )                     \
           if ( (__idx) == (__n)++ ) break;                             \
    }while(0); __head; })                                                

  typedef unsigned long  __usize_t;

  struct  __sll_hash_table {  
     single_ll_t**   table_p;
     __usize_t      length;
  };

  typedef struct  __sll_hash_table  sll_hash_tbl_t;

  #define  hash_table      hash_tbl_p->table_p            
  #define  hash_length     hash_tbl_p->length

  extern __inline__ void  sll_hash_tbl_reset( sll_hash_tbl_t *hash_tbl_p, int is_init )
     {
	  if ( hash_tbl_p ) 
	    {
 	      if( !is_init && hash_table && hash_length > 0 ) free(hash_table);      
              hash_table  = NULL;
	      hash_length = 0UL;
	    }
	  }

  extern __inline__ int sll_hash_tbl_insert ( sll_hash_tbl_t *hash_tbl_p, single_ll_t  *head )
	  {
	   if ( hash_tbl_p && head ) {    
	       hash_table = (single_ll_t **)( hash_table != NULL ?  
	                                      realloc ( hash_table, (hash_length+1) * sizeof(single_ll_t*) ) :
	                                      malloc ( sizeof(single_ll_t*) ) );
	     if ( hash_table ) 
	       {
	         *(hash_table+hash_length) = head;
	         hash_length++;  
	         return 0;
	       } 
	   }
	   return -1;
	  }     

extern __inline__ single_ll_t* sll_hash_tbl_get ( sll_hash_tbl_t *hash_tbl_p, unsigned long idx )
	  {
 	     return ( hash_tbl_p == NULL ? NULL :
	                                (idx < hash_length ? *(hash_table+idx) : NULL) );
	  } 

#ifdef __cplusplus 
}
#endif /*__cplusplus*/ 

#endif /*__SINGLE_LLIST_H__*/

