/*****************************************************************
*   main.c is a shell to run GNU assembler examples that may be found at 
*             my web-site rjaan.narod.ru
*               
* Copyright (C) 2009-2015 Andrey Y. Rjavskov(aka rjaan) <rjaan@yandex.ru> 
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software Foundation,
* Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*
*****************************************************************/
#include <stdio.h>
#include <stdlib.h> 
#include <string.h> 
#include <errno.h>
#include <unistd.h>

#include <sys/un.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "parser.h"
#include "runpart.h"

#ifndef __x86_64__
#include "gasrunparts/gasrunparts.h"
#endif /*__x86_64__*/

#include "gnucrunparts/gnucrunparts.h"

static void help_message ( void );

#define EXAMPLE_DIR "/usr/local/share/gasrunparts/examples"

#define RETFAILURE(msg) {      \
        if(msg) printf(msg);   \
        retval = -EXIT_FAILURE; \
        goto terminated;        \
}
#define RETSUCCESS(msg) {      \
        if(msg) printf(msg,retval);   \
}

#define COMPILE_EXAMPLE_WITH_ALL_ERROR_AND_RETCODE_PRINT(exec,sources,args) {           \
    char *cwd_p = runpart_pwd();                         	                \
    if( cwd_p == NULL )                                                         \
      RETFAILURE("error: can\'t get current working directory!")                \
    printf( "Compile \"%s\" in dir \"%s\"\n", sources, cwd_p );                 \
    if ( runpart_exec ( "gcc -g -Wall -Werror -Os -o %s %s",exec,sources) ) \
        RETFAILURE("error: can\'t compile THIS!");                              \
    printf( "Running \"%s\" in dir \"%s %s\"\n", exec, cwd_p, args );           \
    if ( (retval=runpart_exec ( "/bin/sh -c \"./%s %s\"",exec,args)) == 255 )   \
        RETFAILURE("error: can\'t run THIS!")                                   \
    else 								        \
        RETSUCCESS("retcode: %d") }                    

#define COMPILE_EXAMPLE_RETCODE_PRINT(exec,sources,args) {                           \
    char *cwd_p = runpart_pwd();                         	  \
    if( cwd_p == NULL )                                           \
      RETFAILURE("error: can\'t get current working directory!")  \
    printf( "Compile \"%s\" in dir \"%s\"\n", sources, cwd_p );      \
    if ( runpart_exec ( "gcc -g -Wsequence-point -Werror -Os -o %s %s",exec,sources) ) \
        RETFAILURE("error: can\'t compile THIS!");                   \
    printf( "Running \"%s\" in dir \"%s %s\"\n", exec, cwd_p, args );      \
    if ( (retval=runpart_exec ( "/bin/sh -c \"./%s %s\"",exec,args)) == 255 )  \
        RETFAILURE("error: can\'t run THIS!")                               \
    else 								\
        RETSUCCESS("retcode: %d") }                    

#define COMPILE_EXAMPLE(exec,sources) {                           \
    char *cwd_p = runpart_pwd();                         	  \
    if( cwd_p == NULL )                                           \
      RETFAILURE("error: can\'t get current working directory!")  \
    printf( "Compile \"%s\" in dir \"%s\"\n", sources, cwd_p );      \
    if ( runpart_exec ( "gcc -g -Wsequence-point -Werror -Wno-error=unused-variable -o %s %s",exec,sources) ) \
        RETFAILURE("error: can\'t compile THIS!");                   \
    printf( "Running \"%s\" in dir \"%s\"\n", exec, cwd_p );      \
    if ( runpart_exec ( "./%s",exec) == -1 )\
        RETFAILURE("error: can\'t run THIS!");}

/*
#if 0
    if ( runpart_exec ( "gcc -Wsequence-point -Werror -O2 -S -c %s  -o- | less -N",sources) ) \
#endif 
*/

#define ASM_GENCODE(sources) {                                     \
    char *cwd_p = (char *) runpart_pwd();                          \
    if( cwd_p == NULL )                                            \
      RETFAILURE("error: can\'t get current working directory!")   \
    printf( "Generate assembly code \"%s\" in dir \"%s\"\n", sources, cwd_p ); \
   if ( runpart_exec ( "gcc -Wsequence-point -Werror -fno-omit-frame-pointer -c -g -Wa,-a,-ad %s | less -N",sources) ) \
        RETFAILURE("error: can\'t compile THIS!");}

#define ASM_GENCODE_WITH_NOSTORE_FRAME_POINTER(sources) {            \
    char *cwd_p = (char *) runpart_pwd();                          \
    if( cwd_p == NULL )                                            \
      RETFAILURE("error: can\'t get current working directory!")   \
    printf( "Generate assembly code \"%s\" in dir \"%s\"\n", sources, cwd_p ); \
   if ( runpart_exec ( "gcc -Wsequence-point -Werror -fomit-frame-pointer -c -g -Wa,-a,-ad %s | less -N",sources) ) \
        RETFAILURE("error: can\'t compile THIS!");}


#define WAIT_PRESS_KEY_ENTER(npress,phrasep) ({                    \
     char *__phrasep = (phrasep);                                  \
     int __key, __npress = (npress);         		           \
     do{						           \
	printf( "\n== Press key Enter to %s ==", (__phrasep) ? (__phrasep) : "..." ); \
     }while( ((__key)=getc(stdin)) && --(__npress) )              \
	       ;                                                   \
	       __key; })				  	 

int
main ( int argc, char **argv )
{

  int retval = EXIT_SUCCESS;

  if ( (retval=cmdline_opts_parser ( argc, argv )) < 0 ) {
     help_message();
     exit(EXIT_FAILURE);
  }

#ifdef __DEBUG__
  PRINT_VARIABLE(to_show_help);
  PRINT_VARIABLE(to_show_version);
  PRINT_VARIABLE(to_run_gnuc_dwarf_example);	 
#endif /*__DEBUG__*/

  if ( to_show_help ) {
      goto help_label;
  }

  if ( to_show_version ) {
       VERSION_BANNER();
       goto terminated;
  }

  retval = EXIT_SUCCESS;

  if ( to_run_gnuc_lvalue ) {
     COMPILE_EXAMPLE_RETCODE_PRINT("gnuc_lvalue", 
					      EXAMPLE_DIR "/gnuc_lvalue.c", "");
     WAIT_PRESS_KEY_ENTER(1,"continue");
     ASM_GENCODE_WITH_NOSTORE_FRAME_POINTER ( EXAMPLE_DIR "/gnuc_lvalue.c" );     
     goto success_label;
  }else
  if ( to_run_gnuc_ptralias ) {
     COMPILE_EXAMPLE_RETCODE_PRINT("x86-64_ptralias", EXAMPLE_DIR "/x86_64_ptralias.c", "");
     WAIT_PRESS_KEY_ENTER(1,"continue");
     ASM_GENCODE_WITH_NOSTORE_FRAME_POINTER ( EXAMPLE_DIR "/x86_64_ptralias.c" );     
     goto success_label;
  }else
    if ( to_run_x86_64_typealign ) {
     ASM_GENCODE_WITH_NOSTORE_FRAME_POINTER ( EXAMPLE_DIR "/x86_64_typealign.c" );     
     goto success_label;
  }
  else
  if ( to_run_gnuc_noautomatic_array_nofp ) {
     COMPILE_EXAMPLE_RETCODE_PRINT("noautomatic_arrays", EXAMPLE_DIR "/noautomatic_arrays.c", "");
     WAIT_PRESS_KEY_ENTER(1,"continue");
     ASM_GENCODE_WITH_NOSTORE_FRAME_POINTER ( EXAMPLE_DIR "/noautomatic_arrays.c" );     
     goto success_label;
  }
  else
  if (to_run_gnuc_noautomatic_array){
     COMPILE_EXAMPLE_RETCODE_PRINT("noautomatic_arrays", EXAMPLE_DIR "/noautomatic_arrays.c", "");
     WAIT_PRESS_KEY_ENTER(1,"continue");
     ASM_GENCODE ( EXAMPLE_DIR "/noautomatic_arrays.c" );     
     goto success_label;
  }
  else 
  if (to_run_gnuc_automatic_array){
     COMPILE_EXAMPLE_RETCODE_PRINT("automatic_arrays", EXAMPLE_DIR "/automatic_arrays.c", "");
     WAIT_PRESS_KEY_ENTER(1,"continue");
     ASM_GENCODE_WITH_NOSTORE_FRAME_POINTER ( EXAMPLE_DIR "/automatic_arrays.c" );     
     goto success_label;
  }else
#ifdef __FEATURE_USAGE__
  if ( to_run_gnuc_dwarf_example ) {
      printf("\n == try to run dwarf example == \n" );
     ASM_GENCODE( EXAMPLE_DIR "/hello.c");     
     goto success_label;
  }else
#endif /*__FEATURE_USAGE__*/
  if ( to_run_operand_vs_variable) {
     COMPILE_EXAMPLE_RETCODE_PRINT("operand_vs_variable", EXAMPLE_DIR "/operand_vs_variable.c", "240 90");
     WAIT_PRESS_KEY_ENTER(1,"continue");
     ASM_GENCODE_WITH_NOSTORE_FRAME_POINTER( EXAMPLE_DIR "/operand_vs_variable.c");     
     goto success_label;
  }else
  if ( to_run_algthm_single_ll ) {
     COMPILE_EXAMPLE("algthm_single_ll", EXAMPLE_DIR "/algthm_single_ll.c");
     goto success_label; 
  }else   
    if ( to_run_gnuc_vla_use ) {
     ASM_GENCODE( EXAMPLE_DIR "/vla_use.c");     
     goto success_label;
  }else 
    if ( to_run_gnuc_vla_invalid ) {
     COMPILE_EXAMPLE("vla_invalid", EXAMPLE_DIR "/vla_invalid.c");     
    goto success_label;
  }else 
    if ( to_run_gnuc_unsafe_macros_max ) {
     COMPILE_EXAMPLE("unsafe_fmacros_max", EXAMPLE_DIR "/unsafe_fmacros_max.c");     
    goto success_label;
  }else 
  if ( to_run_gnuc_stmt_expr_max ) {
     COMPILE_EXAMPLE("stmt_expr_max", EXAMPLE_DIR "/stmt_expr_max.c");     
    goto success_label;
#ifndef __x86_64__
  }else
  if( to_run_inums_addition ) {
    printf( "Result: %u\n", inums_addition() );
    goto success_label;
  }else
    if( to_run_list_regimm ) {
       printf( "Result:  %u\n", list_regimm() );
     goto success_label;
  }else
    if( to_run_xchgw_intops ) {
      printf( "Result:  0x%04X\n", xchgw_intops() );
     goto success_label;
  }else
    if( to_run_bswap_intops ) {
      printf( "Result:  0x%08X\n", bswap_intops() );
     goto success_label;
  }else
    if( to_run_trans_negative ) {
      char neg_integer = trans_negative();
      printf( "Result:  0x%X(HEX), %d(DEC)\n", neg_integer, neg_integer  );
     goto success_label;
  }else
    if( to_run_trans_positive ) {
      unsigned char neg_integer = trans_positive();
      printf( "Result:  0x%X(HEX), %d(DEC)\n", neg_integer, neg_integer  );
#endif /*__x86_64__*/
      goto success_label;
  }else
    if( to_run_gnuc_trans_negative ) {
      signed char ret_value = gnuc_trans_negative(); 
      printf( "Input value: %d(DEC), result:  %d(DEC)\n", negative_value, ret_value );
      goto success_label;

   }else
    if( to_run_gnuc_trans_bitwisenot ) {
      signed char ret_value = gnuc_trans_bitwisenot(); 
      printf( "Input value: %d(DEC)[%02X(HEX)], result: %d(DEC)[%02X(HEX)]\n", 
                                                                bitwise_value, 
                                                                bitwise_value, 
                                                                ret_value,
                                                                ret_value );
      goto success_label;
   }else
    if( to_run_gnuc_incr_pointer ) {
      integers_t* output_list_p = gnuc_increment_pointer();       

      printf( "Input data:\n\t\t byte: %d,\n\t\t word: %d,\n\t\t quardword: %d\n", number_list_p->byte,
                                                                                   number_list_p->word,
                                                                                   number_list_p->quardword );
      printf( "Result data:\n\t\t byte: %d,\n\t\t word: %d,\n\t\t quardword: %d\n", output_list_p->byte,
                                                                                    output_list_p->word,
                                                                                    output_list_p->quardword );
      printf( "Сhanging address on: %d\n", output_list_p-number_list_p );
      goto success_label;
   }else
     if( to_run_gnuc_decr_pointer ) {
       integers_t* output_list_p = gnuc_decrement_pointer();       

       printf( "Input data:\n\t\t byte: %d,\n\t\t word: %d,\n\t\t quardword: %d\n", number_list_p->byte,
                                                                                    number_list_p->word,
                                                                                    number_list_p->quardword );
       printf( "Result data:\n\t\t byte: %d,\n\t\t word: %d,\n\t\t quardword: %d\n", output_list_p->byte,
                                                                                    output_list_p->word,
                                                                                    output_list_p->quardword );
       printf( "Сhanging address on: %d\n", output_list_p-number_list_p );
      goto success_label;
  }else{
    goto help_label;
  }

terminated:
  if( retval == -EXIT_FAILURE ) puts ("\nFailure");
  else puts ("\nSuccess");
  return retval;

help_label:
   help_message();
 
success_label:
   puts ("done.");
   return retval;
}

static void help_message ( void )
{
   printf ("%s - \
runs one of assambly functions of GAS examples\n", PACKAGE_STRING );
  printf ("Usage: %s [OPTION] <EXAMPLES>\n", PACKAGE );
  printf ("\
Options:\n\
  -h, --help                 displays this help and exit\n\
  -V, --version              outputs version information and exit\n\
Examples:\n" );
#ifndef __x86_64__
printf ("\
GNU AS\n\
   inums_addition            the addition operation with two operands\n\
   list_regimm               list some usage variants of common registers with applying immediate operands\n\
   bswap_intops              provides the byte change in 32th integers\n\
   xchgw_intops              provides the byte change in 16th integers\n\
   trans_negative            shows negative number -113 in hexadecimal and decimal notations\n\
   trans_positive            shows negative number -113 in hexadecimal and decimal notations\n");
#endif /*__x86_64__*/
printf ("\
GNU C & AS\n\
x86_64_typealign                an example shows why the alignment of data items, in architecture x86-64,are\n\
                                multiples of four, eight or 16 bytes\n\
GNU C are\n\
   gnuc_trans_negative          illustrates the usage of operator \'-\' that trans positive number 113 to \n\
                                two’s complement negation.\n\
   gnuc_trans_bitwisenot        illustrates the usage of operator \'~\' that flips all of the bits\n\
   gnuc_pointer_increment       runs the GNU C example of usage the operation increment of the pointer\n\
   gnuc_stmt_expr_max           example using the expression statement (example of building compiled and executed code)\n\
   gnuc_unsafe_macros_max       why like-function macros may be unsafe (example of building uncompiled code)\n\
   gnuc_vla_invalid             why VLA can\'t be initialized\n\
   gnuc_vla_use                 use VLA in an assembly code\n\
   gnuc_automatic_array         an example of automatic storage object\n\
   gnuc_noautomatic_array       an example of noautomatic storage object (to store the frame pointer)\n\
   gnuc_noautomatic_array_nofp  an example of noautomatic storage object (does not store the frame pointer)\n\
   gnuc_pointer_aliasing        an example of pointer aliasing\n\
   gnuc_locator_value           an example of locator value(lvalue)\n\
\n \
GNU C Algorithms are\n\
   algthm_single_ll          Another of many an implementation of singly linked list\n\
   operand_vs_variable       run simple task is illustrated a difference between    \n\
                             operands and variables\n\n");
 }

/*eof*/

