/* 
   File parser.c contains some functions to parse options of command line.  
 
   Copyright (C) 2015  Andrey Y. Rzhavskov  <mailto:rjaan@yandex.ru>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Affero General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 #include "parser.h"

 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <stdarg.h>
 #include <time.h>

 #include <sys/types.h>
 #include <sys/param.h>

 #include <unistd.h>

 int to_show_help;
 int to_show_version;
 int to_run_inums_addition;
 int to_run_list_regimm;
 int to_run_xchgw_intops;
 int to_run_bswap_intops;
 int to_run_trans_negative;
 int to_run_trans_positive;
 int to_run_gnuc_trans_negative;
 int to_run_gnuc_trans_bitwisenot;
 int to_run_gnuc_incr_pointer; 
 int to_run_gnuc_decr_pointer;
 int to_run_gnuc_stmt_expr_max;
 int to_run_gnuc_unsafe_macros_max;
 int to_run_gnuc_vla_invalid;
 int to_run_gnuc_vla_use;
 int to_run_algthm_single_ll;
 int to_run_operand_vs_variable; 
 int to_run_gnuc_dwarf_example;
 int to_run_gnuc_automatic_array;
 int to_run_gnuc_noautomatic_array;
 int to_run_gnuc_noautomatic_array_nofp;
 int to_run_x86_64_typealign;
 int to_run_gnuc_ptralias;
 int to_run_gnuc_lvalue;

 typedef struct jump_table_struct {

	const char *name; /* a name of long option of command-line */	
        const void *jump; /* a pointer on constant of jump-label */
	int        *extvarp; /* an external variable */

 } jump_table_struct_t;

static int compare_jump_table_structs(const void *ch_p, const void *array_p){   

   const jump_table_struct_t*  a = (const jump_table_struct_t*)ch_p;
   const jump_table_struct_t*  b = (const jump_table_struct_t*)array_p;

   int n = strcmp ( a->name, b->name ); 

   if( b && b->extvarp )
     {
        *(b->extvarp) = ( n ? 0 : 1);
     }	

   return n;	
}


static int globalvar_by_get_option_name ( const char buf[] )
{
 int oneval = 1;

 jump_table_struct_t      findme = { buf, NULL, NULL };
 jump_table_struct_t	 *found; /* a jump found labels */
 
static jump_table_struct_t jump_table[] = { /* the table of jump labels */
	{ "--help",              &&case_longHelp,    &to_show_help  },    /* long option --help */
	{ "--version",           &&case_longVersion, &to_show_version },  /* long option --version */
	{ "-h",                  &&case_longHelp,    &to_show_help },  /* short option '-h'  as same long option --help */
	{ "-V",                  &&case_longVersion, &to_show_version },  /* short option '-V' as same long option --version */
        { "inums_addition",      &&case_inumsAddition,     &to_run_inums_addition },  /* argument 'inums_addition' */
        { "list_regimm",         &&case_listRegimm,        &to_run_list_regimm },   /* argument 'list_regimm' */
        { "xchgw_intops",        &&case_xchgwIntOps,       &to_run_xchgw_intops }, /* argument 'xchgw_intops' */
        { "bswap_intops",        &&case_bswapIntOps,       &to_run_bswap_intops }, /* argument 'bswap_intops' */
        { "trans_negative",      &&case_transNegative,     &to_run_trans_negative }, /* argument 'trans_negative' */
        { "trans_positive",      &&case_transPositive,     &to_run_trans_positive }, /* argument 'trans_positive' */
        { "x86_64_typealign",    &&case_x86_64_typealign,  &to_run_x86_64_typealign }, /* argument 'x86_64_typealign' */ 
        { "gnuc_trans_negative", &&case_GNUC_transNegative,&to_run_gnuc_trans_negative }, /* argument 'gnuc_trans_negative' */
        { "gnuc_trans_bitwisenot", &&case_GNUC_transBitwiseNot,&to_run_gnuc_trans_bitwisenot }, /* argument 'gnuc_trans_bitwisenot' */
        { "gnuc_pointer_increment", &&case_GNUC_incrPointer,&to_run_gnuc_incr_pointer }, /* argument 'gnuc_pointer_increment' */        
        { "gnuc_pointer_decrement", &&case_GNUC_decrPointer,&to_run_gnuc_decr_pointer }, /* argument 'gnuc_pointer_increment' */        
        { "gnuc_stmt_expr_max"    , &&case_GNUC_stmtExprMax,&to_run_gnuc_stmt_expr_max }, /* argument 'gnuc_stmt_expr_max' */
        { "gnuc_unsafe_macros_max", &&case_GNUC_unsafeMacrosMax,&to_run_gnuc_stmt_expr_max }, /* argument 'gnuc_unsafe_macros_max' */
        { "gnuc_automatic_array",   &&case_GNUC_AutoArray, &to_run_gnuc_automatic_array }, /* argument 'gnuc_automatic_array' */
        { "gnuc_noautomatic_array", &&case_GNUC_NoAutoArray, &to_run_gnuc_noautomatic_array }, /* argument 'gnuc_noautomatic_array' */
        { "gnuc_vla_invalid",       &&case_GNUC_VLAinvalid, &to_run_gnuc_vla_invalid }, /* argument 'gnuc_vla_invalid' */
        { "gnuc_vla_use",           &&case_GNUC_VLAuse, &to_run_gnuc_vla_use }, /* argument 'gnuc_vla_use' */
        { "algthm_single_ll",       &&case_ALGTHM_LL, &to_run_algthm_single_ll   }, /* argument 'algthm_single_ll' */
        { "operand_vs_variable",    &&case_OPERAND_VS_VARIABLE,&to_run_operand_vs_variable }, /* argument 'operand_vs_variable' */
        { "gnuc_dwarf_example",     &&case_DWARF_EXAMPLE, &to_run_gnuc_dwarf_example }, /* argument 'dwarf_example' */
        { "gnuc_noautomatic_array_nofp", &&case_Success,  &to_run_gnuc_noautomatic_array_nofp },
	{ "gnuc_pointer_aliasing",  &&case_GNUC_ptralias, &to_run_gnuc_ptralias },  /* argument 'gnuc_ptralias' */
        { "gnuc_locator_value"   ,  &&case_GNUC_lvalue,   &to_run_gnuc_lvalue }   /* argument 'gnuc_locator_value' */
 };

 const int jump_table_count =  /* calculating a number of jump labels into the table   */
			   sizeof(jump_table)/sizeof(jump_table_struct_t);

 int ncount = jump_table_count;

/* pre-sort jump label*/
 qsort (jump_table, jump_table_count,
      		      sizeof(jump_table_struct_t), compare_jump_table_structs );
/* search jump label*/
 found = bsearch ( &findme, jump_table, jump_table_count,
      		      sizeof(jump_table_struct_t), compare_jump_table_structs );
 if(!found) {
   goto case_Failure; 
 }

 if( found->jump && found->extvarp ) {
       *(found->extvarp) = 1;
#ifdef __DEBUG__
        printf("\n == found %s option, value: %d == \n", found->name, *found->extvarp );
#endif /*__DEBUG__*/
       goto *(found->jump);
 }	

case_OPERAND_VS_VARIABLE:
	goto case_Success;
case_ALGTHM_LL:
	goto case_Success;
case_GNUC_VLAuse:
	goto case_Success;
case_GNUC_VLAinvalid:
	goto case_Success;
case_GNUC_stmtExprMax:
	goto case_Success;
case_GNUC_unsafeMacrosMax:
	goto case_Success;
case_longHelp:
	goto case_Success;
case_longVersion:
	goto case_Success;
case_inumsAddition:
	goto case_Success;
case_listRegimm:
	goto case_Success;
case_xchgwIntOps:
        goto case_Success;
case_bswapIntOps:
        goto case_Success;
case_transNegative:
        goto case_Success;
case_transPositive:
        goto case_Success;
case_x86_64_typealign:
        goto case_Success;
case_GNUC_incrPointer:
        goto case_Success;
case_GNUC_decrPointer:
        goto case_Success;
case_GNUC_transBitwiseNot:
        goto case_Success;
case_GNUC_transNegative:
        goto case_Success;
case_DWARF_EXAMPLE:
        goto case_Success;
case_GNUC_AutoArray:
        goto case_Success;
case_GNUC_NoAutoArrayWithoutFP:
        goto case_Success;
case_GNUC_NoAutoArray:
        goto case_Success;
case_GNUC_ptralias:
        goto case_Success;
case_GNUC_lvalue:
        goto case_Success;

case_Failure:
	return -1;
case_Success:
	return 0;

}

#define OPTION_NOTFOUND()\
 printf ( "%s: invalid option -- \'%s\'\n\n", *argv, *p_argv_go )

#define ARGUMENT_NOTFOUND(NAME)\
 printf ( "%s: \'%s\' option required argument\n", *argv, NAME )


/*
   cmdline_opts_parser() parses  command-line options  that were  passed    into 
			 2nd argument and initializes global variables that were
			 passed since 3rd argument.
 	
		     	 On success, this function  returns  a number of options 
		         were recognized. Otherwise, it returns negative value.
*/
int cmdline_opts_parser ( const int argc, char * const *argv ) 
 {
    va_list  ap;
    int      n_parsed_opts  	   = 0;      
    int      to_go          	   = argc-1;   
    char     * const *p_argv_go    = argv+1; 
    char     is_wait_arg    	   = 0;
   
    for ( ; to_go && *p_argv_go; p_argv_go++, to_go-- ) {
	int rval = globalvar_by_get_option_name ( *p_argv_go );	
	if( rval < 0 ) {
	     OPTION_NOTFOUND();
	     n_parsed_opts = -1;
	     goto out;
	 }
	else
	 if( !rval ) 
	 {
	   return 1;
	 }
         n_parsed_opts++;
    }	   
    return n_parsed_opts;
out:
    return  -EXIT_FAILURE;
 }

/*eof*/

